@@ -1,5 +1,9 @@
 #!/usr/bin/env python3
 
+r"""
+A set of utilities for data & tasks loading, preprocessing, and sampling.
+"""
+
 from . import transforms
 from .meta_dataset import MetaDataset
 from .task_dataset import TaskDataset, DataDescription
