@@ -1,7 +1,6 @@
 #!/usr/bin/env python3
 
 import copy
-
 import torch
 
 
@@ -49,7 +48,7 @@ def clone_parameters(param_list):
     return [p.clone() for p in param_list]
 
 
-def clone_module(module):
+def clone_module(module, memo=None):
     """
 
     [[Source]](https://github.com/learnables/learn2learn/blob/master/learn2learn/utils.py)
@@ -92,9 +91,17 @@ def clone_module(module):
     #       clone = recursive_shallow_copy(model)
     #       clone._apply(lambda t: t.clone())
 
+    if memo is None:
+        # Maps original data_ptr to the cloned tensor.
+        # Useful when a Module uses parameters from another Module; see:
+        # https://github.com/learnables/learn2learn/issues/174
+        memo = {}
+
     # First, create a copy of the module.
     # Adapted from:
     # https://github.com/pytorch/pytorch/blob/65bad41cbec096aa767b3752843eddebf845726f/torch/nn/modules/module.py#L1171
+    if not isinstance(module, torch.nn.Module):
+        return module
     clone = module.__new__(type(module))
     clone.__dict__ = module.__dict__.copy()
     clone._parameters = clone._parameters.copy()
@@ -105,20 +112,36 @@ def clone_module(module):
     if hasattr(clone, '_parameters'):
         for param_key in module._parameters:
             if module._parameters[param_key] is not None:
-                cloned = module._parameters[param_key].clone()
-                clone._parameters[param_key] = cloned
+                param = module._parameters[param_key]
+                param_ptr = param.data_ptr
+                if param_ptr in memo:
+                    clone._parameters[param_key] = memo[param_ptr]
+                else:
+                    cloned = param.clone()
+                    clone._parameters[param_key] = cloned
+                    memo[param_ptr] = cloned
 
     # Third, handle the buffers if necessary
     if hasattr(clone, '_buffers'):
         for buffer_key in module._buffers:
             if clone._buffers[buffer_key] is not None and \
                     clone._buffers[buffer_key].requires_grad:
-                clone._buffers[buffer_key] = module._buffers[buffer_key].clone()
+                buff = module._buffers[buffer_key]
+                buff_ptr = buff.data_ptr
+                if buff_ptr in memo:
+                    clone._buffers[buffer_key] = memo[buff_ptr]
+                else:
+                    cloned = buff.clone()
+                    clone._buffers[buffer_key] = cloned
+                    memo[param_ptr] = cloned
 
     # Then, recurse for each submodule
     if hasattr(clone, '_modules'):
         for module_key in clone._modules:
-            clone._modules[module_key] = clone_module(module._modules[module_key])
+            clone._modules[module_key] = clone_module(
+                module._modules[module_key],
+                memo=memo,
+            )
 
     # Finally, rebuild the flattened parameters for RNNs
     # See this issue for more details:
@@ -152,6 +175,8 @@ def detach_module(module):
     error.backward()  # Gradients are back-propagate on clone, not net.
     ~~~
     """
+    if not isinstance(module, torch.nn.Module):
+        return
     # First, re-write all parameters
     for param_key in module._parameters:
         if module._parameters[param_key] is not None:
@@ -174,12 +199,12 @@ def clone_distribution(dist):
 
     for param_key in clone.__dict__:
         item = clone.__dict__[param_key]
-        if isinstance(item, th.Tensor):
+        if isinstance(item, torch.Tensor):
             if item.requires_grad:
                 clone.__dict__[param_key] = dist.__dict__[param_key].clone()
-        elif isinstance(item, th.nn.Module):
+        elif isinstance(item, torch.nn.Module):
             clone.__dict__[param_key] = clone_module(dist.__dict__[param_key])
-        elif isinstance(item, th.Distribution):
+        elif isinstance(item, torch.Distribution):
             clone.__dict__[param_key] = clone_distribution(dist.__dict__[param_key])
 
     return clone
@@ -189,11 +214,77 @@ def detach_distribution(dist):
     # TODO: This function was never tested.
     for param_key in dist.__dict__:
         item = dist.__dict__[param_key]
-        if isinstance(item, th.Tensor):
+        if isinstance(item, torch.Tensor):
             if item.requires_grad:
                 dist.__dict__[param_key] = dist.__dict__[param_key].detach()
-        elif isinstance(item, th.nn.Module):
+        elif isinstance(item, torch.nn.Module):
             dist.__dict__[param_key] = detach_module(dist.__dict__[param_key])
-        elif isinstance(item, th.Distribution):
+        elif isinstance(item, torch.Distribution):
             dist.__dict__[param_key] = detach_distribution(dist.__dict__[param_key])
     return dist
+
+
+def update_module(module, updates=None):
+    r"""
+    [[Source]](https://github.com/learnables/learn2learn/blob/master/learn2learn/utils.py)
+
+    **Description**
+
+    Updates the parameters of a module in-place, in a way that preserves differentiability.
+
+    The parameters of the module are swapped with their update values, according to:
+    \[
+    p \gets p + u,
+    \]
+    where \(p\) is the parameter, and \(u\) is its corresponding update.
+
+
+    **Arguments**
+
+    * **module** (Module) - The module to update.
+    * **updates** (list, *optional*, default=None) - A list of gradients for each parameter
+        of the model. If None, will use the tensors in .update attributes.
+
+    **Example**
+    ~~~python
+    error = loss(model(X), y)
+    grads = torch.autograd.grad(
+        error,
+        model.parameters(),
+        create_graph=True,
+    )
+    updates = [-lr * g for g in grads]
+    l2l.update_module(model, updates=updates)
+    ~~~
+    """
+    if updates is not None:
+        params = list(module.parameters())
+        if not len(updates) == len(list(params)):
+            msg = 'WARNING:update_module(): Parameters and updates have different length. ('
+            msg += str(len(params)) + ' vs ' + str(len(updates)) + ')'
+            print(msg)
+        for p, g in zip(params, updates):
+            p.update = g
+
+    # Update the params
+    for param_key in module._parameters:
+        p = module._parameters[param_key]
+        if p is not None and hasattr(p, 'update') and p.update is not None:
+            module._parameters[param_key] = p + p.update
+
+    # Second, handle the buffers if necessary
+    for buffer_key in module._buffers:
+        buff = module._buffers[buffer_key]
+        if buff is not None and hasattr(buff, 'update') and buff.update is not None:
+            module._buffers[buffer_key] = buff + buff.update
+
+    # Then, recurse for each submodule
+    for module_key in module._modules:
+        module._modules[module_key] = update_module(module._modules[module_key],
+                                                    updates=None)
+
+    # Finally, rebuild the flattened parameters for RNNs
+    # See this issue for more details:
+    # https://github.com/learnables/learn2learn/issues/139
+    module._apply(lambda x: x)
+    return module
